#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019 Ina (Zohra Rezgui & David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import cv2
import numpy as np
from matplotlib import pyplot as plt
from .rect import Rect


def _angle_between_2_points(p1, p2):
    x1, y1 = p1
    x2, y2 = p2
    if x1 != x2:
        tan = (y2 - y1) / (x2 - x1)
        return np.degrees(np.arctan(tan))
    elif y2 > y1:
        return np.degrees(np.pi / 2)

    else:
        return np.degrees(-np.pi / 2)


def alignCrop(frame, bb, left_eye, right_eye, verbose=False):
    """
    Rotate image such as the eyes lie on a horizontal line

    Parameters
    ----------
    frame : numpy.ndarray (height,with, 3)
        RGB image data
    bb : (x1, y1, x2, y2)
    left_eye: x, y
    right_eye: x, y

    Returns
    -------
    rotated_frame : numpy.ndarray (height,with, 3)
        RGB rotated and cropped image data
    """
    w = int(bb[2] - bb[0])
    h = int(bb[3] - bb[1])

    angle = _angle_between_2_points(left_eye, right_eye)
    xc = (left_eye[0] + right_eye[0]) / 2
    yc = (left_eye[1] + right_eye[1]) / 2
    M = cv2.getRotationMatrix2D((xc, yc), angle, 1)
    M += np.array([[0, 0, -bb[0]], [0, 0, -bb[1]]])


    rotated_frame = cv2.warpAffine(frame, M, (w, h), flags=cv2.INTER_CUBIC)

    if verbose:
        print('after rotation & crop')
        plt.imshow(rotated_frame)
        plt.show()
    return rotated_frame



def preprocess_face(frame, detection, squarify, bbox_scale, face_alignment, output_shape, verbose=False):
    """
    Apply preprocessing pipeline to a detected face and returns the
    corresponding image with the following optional processings
    Parameters
    ----------
    frame : numpy nd.array
        RGB image data
    detection : a Detection instance containing a field 'bbox' of type Rect
        corresponding to the detected face bounding box
        if not None, the provided bounding box will be used, else the
        whole image will be used
    squarify: boolean
        if set to True, the bounding box will be extented to be the
        smallest square containing the bounding box. This avoids distortion
        if faces are resized in a latter stage
    bbox_scale: float
        resize the bounding box to consider larger (bbox_scale > 1) or
        smaller (bbox_scale < 1) areas around faces. If set to 1, the
        original bounding box is used
    face_alignment: instance of class defined in face_alignment.py or None
        if set to None, do nothing
        if set to an instance of class defined in face_alignment.py
        (for instance Dlib68FaceAlignment) face will be rotated based on the
        estimation of facial landmarks such the eyes lie on a horizontal line
    output_shape: (width, height) or None
        if not None, face will be resized to the provided output shape
    Returns
    -------
    frame: np.array RGB image data
    bbox: up-to-data bounding box after the proprocessing pipeline

    """

    (frame_h, frame_w, _) = frame.shape

    # if no bounding box is provided, use the whole image
    if detection is None:
        bbox = Rect(0, 0, frame_w, frame_h)
    elif isinstance(detection, Rect):
        bbox = detection
    else:
        bbox = detection.bbox

    if face_alignment is not None:
        left_eye, right_eye = face_alignment(frame, bbox)

    # if True, extend the bounding box to the smallest square containing
    # the orignal bounding box
    if squarify:
        bbox = bbox.square

    # perform bounding box scaling to march larger/smaller areas around
    # the detected face
    bbox = bbox.scale(bbox_scale)

    bbox = bbox.to_int()


    # performs face alignment based on facial landmark detection
    if face_alignment is not None:
        frame = alignCrop(frame, bbox, left_eye, right_eye, verbose=verbose)
    else:
        # crop image to the bounding box
	# TODO: replace by a wrap affine => management of out of frame
        frame = frame[bbox.y1:bbox.y2, bbox.x1:bbox.x2]

    # resize image to the required output shape
    if output_shape is not None:
        frame = cv2.resize(frame, output_shape)

    if verbose:
        print('resulting image')
        plt.imshow(frame)
        plt.show()

    return frame, bbox



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2022 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Here are defined the common pieces of code used by command line argument programs

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import inaFaceAnalyzer as ifa
from inaFaceAnalyzer.face_classifier import faceclassifier_factory
from inaFaceAnalyzer.face_detector import facedetection_factory

epilog = '''
If you are using inaFaceAnalyzer in your research, please cite
the current version number used (%s) together with a reference to the following
paper: David Doukhan and Thomas Petit (2022). inaFaceAnalyzer: a Python toolbox
for large-scale face-based description of gender representation in media with
limited gender, racial and age biases. Submitted to JOSS - The journal of Open
Source Software (submission in progress).
''' % ifa.__version__


def new_parser(description):
    parser = ArgumentParser(description=description,
                            epilog=epilog,
                            formatter_class= ArgumentDefaultsHelpFormatter)# ,
                            #add_help=False)
    return parser

hfps = '''Amount of video frames to be processed per second.
Remaining frames will be skipped.
If not provided, all video frames will be processed (generally between 25 and 30 per seconds).
Lower FPS values result in faster processing time'''
def add_fps(parser):
    parser.add_argument('--fps', default=None, type=float, help=hfps)

# hkeyframes = '''Face detection and analysis from video limited to video key frames.
# Allows fastest video analysis time associated to a summary with
# non uniform frame sampling rate. Incompatible with the --fps, --ass_subtitle_export or --mp4_export arguments.'''
# def _add_keyframes(parser):
#     parser.add_argument('--keyframes', action='store_true', help=hkeyframes)

# def add_framerate(parser):
#     # keyframes and fps arguments are mutually exclusive
#     group = parser.add_mutually_exclusive_group()
#     _add_fps(group)
#     _add_keyframes(group)

htracking = '''Face tracking works jointly with face detection systems :
a face should be first detected before being tracked.
Face detection will be performed one every DETECT_PERIOD frames and allow to
detect new faces, or faces that were lost due to occlusions.
Face tracking is performed on every frame.
Default DETECT_PERIOD (1) is the most costly and the most robust: face detection will
be performed for every video frames together with face tracking allowing to smooth results.
Larger DETECT_PERIOD are efficient as long as analysis FPS is high, and users
should define a proper trade-off between DETECT_PERIOD and the FPS.
High DETECT_PERIOD and low FPS will results in missed faces due to occlusions
or fast changes in face position.
'''
def add_tracking(parser):
    # TODO: add tracking threshold in the options
    tg = parser.add_argument_group('Arguments specific to "videotracking" engine')
    tg.add_argument('--detect_period', type=int, default=1, dest='detect_period', help=htracking)

def add_batchsize(parser):
    parser.add_argument('--batch_size', default=32, type=int,
                help = '''GPU batch size. Larger values allow faster processings, but require more GPU memory.
                Default 32 value used is fine for a Laptop Quadro T2000 Mobile GPU with 4 Gb memory.''')

def engine_factory(args):
    """
    Instantiante classifier, face detector and analysis engine from command
    line arguments
    Parameters
    ----------
    args : argparse.Namespace
        result of argparse.parse()

    Returns
    -------
    Analysis engine

    """

    bs = args.batch_size
    engine = args.engine

    # classifier constructor
    classifier = faceclassifier_factory(args)

    if args.engine == 'preprocessed_image':
        # TODO: make an engine class
        return classifier

    # face detection constructor
    print('engine factory args', args)
    detector = facedetection_factory(args)

    if engine == 'image':
        return ifa.ImageAnalyzer(detector, classifier, bs)
    if engine == 'video':
        return ifa.VideoAnalyzer(detector, classifier, bs)
    if engine == 'videotracking':
        return ifa.VideoTracking(args.detect_period, detector, classifier, bs)
    if engine == 'videokeyframes':
        engine = ifa.VideoKeyframes(detector, classifier, bs)
    raise NotImplementedError()


#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from typing import NamedTuple
import dlib
import numpy as np

class Rect(NamedTuple):
    x1 : float # left
    y1 : float # top
    x2 : float # right
    y2 : float # bottom
    @property
    def w(self):
        """ returns Rect width """
        return self.x2 - self.x1
    @property
    def h(self):
        """ returns Rect height """
        return self.y2 - self.y1
    @property
    def center(self):
        """ returns center (x,y) of Rect (x1, y1, x2, y2) """
        x1, y1, x2, y2 = self
        return ((x1 + x2) / 2), ((y1 + y2) / 2)
    @property
    def area(self):
        return self.w * self.h
    @property
    def max_dim_len(self):
        """ Return max of Rect width and height """
        return max(self.h, self.w)
    def transpose(self, x, y):
        """ Returns a transposed Rect"""
        x1, y1, x2, y2 = self
        return Rect(x1 + x, y1 + y, x2 + x, y2 + y)
    def mult(self, x, y):
        x1, y1, x2, y2 = self
        return Rect(x1 * x, y1 * y, x2 * x, y2 * y)

    def intersect(self, r):
        x1, y1, x2, y2 = self
        ret = Rect(max(x1, r.x1), max(y1, r.y1), min(x2, r.x2), min(y2, r.y2))
        if ret.h <= 0 or ret.w <= 0:
            return Rect(0,0,0,0)
        return ret

    def iou(self, r):
        inter = self.intersect(r).area
        union = self.area + r.area - inter
        return inter / union

    def __contains__(self, point):
        x1, y1, x2, y2 = self
        x, y = point
        return x >= x1 and x <= x2 and y >= y1 and y <= y2

    def to_int(self):
        return Rect(*[int(round(e)) for e in self])

    @staticmethod
    def from_dlib(x):
        return Rect(x.left(), x.top(), x.right(), x.bottom())

    def to_dlibInt(self):
        return dlib.rectangle(*[e for e in self.to_int()])

    def to_dlibFloat(self):
        return dlib.drectangle(*self)

    @property
    def square(self):
        """ returns the smallest square containing the rectangle"""
        offset = self.max_dim_len / 2
        xc, yc = self.center
        return Rect(xc - offset, yc - offset, xc + offset, yc + offset)

    def scale(self, scale_prct):
        """ scale Rectangle according to scale percentage scale_prct"""
        w, h = (self.w, self.h)
        x1, y1, x2, y2 = self
        xdiff = (w * scale_prct - w) / 2
        ydiff = (h * scale_prct -h) / 2
        return Rect(x1 - xdiff, y1 - ydiff, x2 + xdiff, y2 + ydiff)




#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan & Zohra Rezgui- http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numpy as np
import pandas as pd
import numbers
from abc import ABC, abstractmethod
import tensorflow
from tensorflow import keras
from tensorflow.keras.preprocessing.image import img_to_array

import inaFaceAnalyzer.keras_vggface_patch as keras_vggface
from .svm_utils import svm_load
from .opencv_utils import imread_rgb, disp_frame
from .remote_utils import get_remote


class FaceClassifier(ABC):
    """
    Abstract class to be implemented by face classifiers
    """

    # The 3 properties bellow (input_shape, bbox_scale, bbox2square) are
    # currently common to all implemented face classifiers
    # they provide information on the face preprocessing steps used for
    # training the classification models
    # in future, they may be defined separately for each classifier using
    # abstract properties

    # input image dimensions required by the classifier (width, height, depth)
    input_shape = (224, 224, 3)

    # implemented classifiers are optimized for a given scale factor to be
    # applied on face bounding boxes to be defined here
    bbox_scale = 1.1

    # implemented face classifiers may require a preprocessing step consisting
    # to extend the face bounding box such as the resulting box is the smallest
    # square containing the detected face
    bbox2square = True


    @abstractmethod
    def list2batch(self, limg): pass

    @abstractmethod
    def inference(self, bfeats): pass

    @abstractmethod
    def decisionfunction2labels(self, df): pass

    @property
    def output_cols(self):
        if not hasattr(self, '_output_cols'):
            fake_input = [np.zeros(self.input_shape)]
            self._output_cols = list(self(fake_input, False).columns)
        return self._output_cols

    def average_results(self, df):
        if len(df) == 0:
            for c in self.output_cols:
                df[c] = []

        cols = [e for e in df.columns if e.endswith('_decfunc')]
        gbm = df.groupby('face_id')[cols].mean()
        gbm = self.decisionfunction2labels(gbm)

        return df.join(gbm, on='face_id', rsuffix='_avg')

    # TODO : Keras trick for async READ ?
    # bench execution time : time spent in read/exce . CPU vs GPU
    # TODO: add progress bar with verbose option
    def preprocessed_img_list(self, lfiles, return_features=False, batch_len=32):
        """
        Performs classification on a list of preprocessed face images
        Preprocessed face images are assumed to contain a single face which is
        already detected, cropped, aligned and scaled to classifier's input
        dimensions (for now: 224*224 pixels)

        Parameters
        ----------
        lfiles : list
            list of image paths
        return_features : bool, optional
            Not yet implemented. The default is False.
        batch_len : input, optional
            Larger batch sizes results in faster processing times.
            Batch lenght is dependent on available GPU memory
            The default is 32 (suitable for a laptop GPU).

        Returns
        -------
        df : pandas.DataFrame
            a pandas dataframe with one record for each input image
        """
        assert len(lfiles) > 0

        if return_features:
            raise NotImplementedError()

        lbatchret = []

        for i in range(0, len(lfiles), batch_len):
            xbatch = [imread_rgb(e) for e in lfiles[i:(i+batch_len)]]

            lbatchret.append(self(xbatch, False)) # to change when return features will be managed

        df = pd.concat(lbatchret).reset_index(drop=True)
        df.insert(0, 'filename', lfiles)
        return df

    def __call__(self, limg, return_features, verbose=False):
        """
        Classify a list of images
        images are supposed to be preprocessed faces: aligned, cropped
        Parameters
        ----------
        limg : list of images, a single image can also be used

        Returns
        -------
        feats :
            face features used as input to the final classifier
        label : str
            f for female, m for male
        decision_value : float
            decision function value (negative for female, positive for male)
        """

        if isinstance(limg, list):
            islist = True
        else:
            islist = False
            limg = [limg]

        assert np.all([e.shape == self.input_shape for e in limg])
        batch_ret_feats, batch_ret_preds = self.inference(self.list2batch(limg))
        batch_ret_preds = self.decisionfunction2labels(batch_ret_preds)

        if verbose:
            for img, pred in zip(limg, batch_ret_preds.itertuples(index=False, name='FaceClassifierResult')):
                disp_frame(img)
                print('prediction', pred)

        if islist:
            if return_features:
                return batch_ret_feats, batch_ret_preds
            return batch_ret_preds

        ret = next(batch_ret_preds.itertuples(index=False, name='FaceClassifierResult'))
        if return_features:
            return  batch_ret_feats[0, :], ret
        return ret

class Resnet50FairFace(FaceClassifier):

    def __init__(self):
        m = keras.models.load_model(get_remote('keras_resnet50_fairface.h5'), compile=False)
        self.model = tensorflow.keras.Model(inputs=m.inputs, outputs=m.outputs + [m.layers[-3].output])

    def list2batch(self, limg):
        x = np.concatenate([np.expand_dims(img_to_array(e), axis=0) for e in limg])
        return tensorflow.keras.applications.resnet50.preprocess_input(x)

    def inference(self, x):
        decisions, feats = self.model.predict(x)
        df = pd.DataFrame(decisions.ravel(), columns=['sex_decfunc'])
        return feats, df

    def decisionfunction2labels(self, df):
        df['sex_label'] = df.sex_decfunc.map(lambda x: 'm' if x > 0 else 'f' )
        return df


def _fairface_agedec2age(age_dec):
    ages = np.array([(0,2), (3,9), (10,19), (20,29), (30,39), (40,49), (50,59), (60,69), (70, 79), (80,99)], dtype=np.float32)
    ages_mean = (np.sum(ages, axis=1) + 1) / 2.
    ages_range = ages[:, 1] - ages[:, 0] +1

    if isinstance(age_dec, numbers.Number):
        age_dec = np.array([age_dec])

    age_dec = np.array(age_dec)
    minval = -.5 + 10**-6
    age_dec[age_dec < minval] = minval
    maxval = 9.5 - 10**-6
    age_dec[age_dec > maxval] = maxval
    idec = np.round(age_dec).astype(np.int32)

    age_label = ages_mean[idec] + (age_dec - idec) * ages_range[idec]
    return age_label




class Resnet50FairFaceGRA(Resnet50FairFace):
    def __init__(self):
        m = keras.models.load_model(get_remote('keras_resnet50_fairface_GRA.h5'), compile=False)
        self.model = tensorflow.keras.Model(inputs=m.inputs, outputs=m.outputs + [m.layers[-5].output])

    def inference(self, x):
        gender, _, age, feats = self.model.predict(x)
        df = pd.DataFrame(zip(gender.ravel(), age.ravel()), columns=['sex_decfunc', 'age_decfunc'])
        return feats, df

    def decisionfunction2labels(self, df):
        df = super().decisionfunction2labels(df)
        df['age_label'] = _fairface_agedec2age(df.age_decfunc)
        return df

class OxfordVggFace(FaceClassifier):
    '''
    OxfordVggFace instances are based on pretrained VGG16 architectures
    pretrained using a triplet loss paradigm allowing to obtain face neural
    representation, that we use to train linear SVM classification systems.

    The approach used is fully described in Zohra Rezgui's internship report at INA:
    Détection et classification de visages pour la description de l’égalité
    femme-homme dans les archives télévisuelles, Higher School of Statistics
    and Information Analysis, University of Carthage, 2019

    This class takes advantage of Refik Can Malli's keras-vggface module,
    providing pretrained VGG16 models
    https://github.com/rcmalli/keras-vggface
    '''

    def __init__(self, hdf5_svm=None):
        # Face feature extractor from aligned and detected faces
        self.vgg_feature_extractor = keras_vggface.VGG16(self.input_shape)
        # SVM trained on VGG neural features
        if hdf5_svm is not None:
            self.gender_svm = svm_load(hdf5_svm)

    def decisionfunction2labels(self, df):
        df['sex_label'] = [self.gender_svm.classes_[1 if x > 0 else 0] for x in df.sex_decfunc]
        return df

    def list2batch(self, limg):
        """
        returns VGG16 Features
        limg is a list of preprocessed images supposed to be aligned and cropped and resized to 224*224
        """
        limg = [np.expand_dims(img_to_array(e[:, :, ::-1]), axis=0) for e in limg]
        x = keras_vggface.preprocess_input(np.concatenate(limg))
        return self.vgg_feature_extractor(x)

    def inference(self, x):
        return np.array(x), pd.DataFrame(self.gender_svm.decision_function(x), columns=['sex_decfunc'])

class Vggface_LSVM_YTF(OxfordVggFace):
    def __init__(self):
        OxfordVggFace.__init__(self, get_remote('svm_ytf_zrezgui.hdf5'))


class Vggface_LSVM_FairFace(OxfordVggFace):
    def __init__(self):
        OxfordVggFace.__init__(self, get_remote('svm_vgg16_fairface.hdf5'))

help = '''face classifier to be used in the analysis:
Resnet50FairFaceGRA predicts age and gender and is the most accurate.
It uses Resnet50 architecture and is trained to predict gender, age and race on FairFace.
After consultation of French CNIL (French data protection authority) and
DDD (French Rights Defender), racial classification layers were erased
from this public distribution in order to prevent their use for non ethical purposes.
These models can however be provided for free after examination of each demand.
Resnet50FairFace only predicts gender, and is trained on FairFace with a Resnet50 architecture.
Vggface_LSVM_YTF predicts only gender. It uses an Oxford VGG 16 neural representation
of faces combined with a linear SVM that was trained on Youtube Faces database
by Zohra Rezgui during her internship at INA. It was used in digital earlier humanities studies.
Vggface_LSVM_Fairface has the same architecture and equivalent performances than
Vggface_LSVM_YTF. Its linear SVM model was trained on FairFace.
'''
choices = ['Resnet50FairFaceGRA', 'Vggface_LSVM_YTF', 'Resnet50FairFace', 'Vggface_LSVM_FairFace']
def faceclassifier_cmdline(parser):
    parser.add_argument ('--classifier', default='Resnet50FairFaceGRA',
                         choices = choices, help = help)

def faceclassifier_factory(args):
    if args.classifier == 'Resnet50FairFaceGRA':
        return Resnet50FairFaceGRA()
    if args.classifier == 'Resnet50FairFace':
        return Resnet50FairFace()
    if args.classifier == 'Vggface_LSVM_FairFace':
        return Vggface_LSVM_FairFace()
    if args.classifier == 'Vggface_LSVM_YTF':
        return Vggface_LSVM_YTF()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2021 Ina (David Doukhan & Zohra Rezgui - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import cv2
import tempfile
import os
import pandas as pd
from Cheetah.Template import Template
import datetime
from .opencv_utils import video_iterator, get_video_properties, analysisFPS2subsamp_coeff


def _hex2rgb(hx):
    return (int(hx[0:2],16), int(hx[2:4],16), int(hx[4:6],16))


def _sec2hmsms(s):
    td = datetime.timedelta(seconds=s)
    h,m,s = str(td).split(':')
    return '%d:%d:%.2f' % (int(h), int(m), float(s))


def _analysis2displaydf(df, fps, subsamp_coeff, text_pat = None, cols=None):
    """
    Convert analysis results to a generic pandas dataframe containing
    formated information to be displayed using export functions defined bellow
    """
    if isinstance(df, str):
        df = pd.read_csv(df)
        df.bbox = df.bbox.map(lambda x: eval(x))

    ret = pd.DataFrame()
    ret['frame'] = df.frame
    ret['bbox'] = df.bbox
    ret[['x1', 'y1', 'x2', 'y2']] = df.apply(lambda x: x.bbox, axis=1, result_type="expand")
    ret['start'] = df.frame.map(lambda x: _sec2hmsms(x / fps))
    ret['stop'] = df.frame.map(lambda x: _sec2hmsms((x + subsamp_coeff) / fps))
    if text_pat is None:
        if 'face_id' in df.columns:
            ret['rgb_color'] = df.sex_label_avg.map(lambda x: '0000FF' if x == 'm' else '00FF00')
            text_pat = 'id: %s'
            cols = ['face_id']
            if 'sex_label_avg' in df.columns:
                text_pat += ' - sex: %s (%.1f)'
                cols += ['sex_label_avg', 'sex_decfunc_avg']
            if 'age_label_avg' in df.columns:
                text_pat += ' - age: %.1f'
                cols += ['age_label_avg']
        else:
            text_pat = 'sex: %s (%.1f)'
            cols = ['sex_label', 'sex_decfunc']
            ret['rgb_color'] = df.sex_label.map(lambda x: '0000FF' if x == 'm' else '00FF00')
            if 'age_label' in df.columns:
                text_pat += '- age: %.1f'
                cols += ['age_label']

    ret['bgr_color'] = ret.rgb_color.map(lambda x: x[4:] + x[2:4] + x[:2])
    ret['text'] = df.apply(lambda x: text_pat % tuple([x[e] for e in cols]), axis=1)
    return ret

def ass_subtitle_export(vid_src, result_df, subtitle_dst, analysis_fps=None):
    """
    Export inaFaceAnalyzer results to an ASS subtitle file allowing to display
    faces bounding boxes and other analysis information
    ASS subtitles can be displayed using VLC or Aegisub

    Parameters
    ----------
    vid_src : str
        path to the input video.
    result_df : str or pandas.DataFrame
        path to the csv or pandas DataFrame corresponding to the analysis of the input video.
    subtitle_dst : str
        output path which will store the resulting subtitle file.
        Must have ass extension
    analysis_fps: numeric or None
        The amount of frames per second which were analyzed
        if set to None, then consider that all video frames were processed
    """


    assert subtitle_dst[-4:].lower() == '.ass', subtitle_dst

    video_props = get_video_properties(vid_src)
    fps, width, height = [video_props[e] for e in ['fps', 'width', 'height']]

    if analysis_fps is None:
        subsamp_coeff = 1
    else:
        subsamp_coeff = analysisFPS2subsamp_coeff(vid_src, analysis_fps)

    displaydf = _analysis2displaydf(result_df, fps, subsamp_coeff)

    p = os.path.dirname(__file__)
    t = Template(file = p + '/template.ass')

    t.height = height
    t.width = width
    t.display_df = displaydf
    # text font size set to 4% video height
    t.text_font_size = int(0.04 * height)

    with open(subtitle_dst, 'wt') as fid:
        print(t, file=fid)


def video_export(vid_src, result_df, vid_dst, analysis_fps=None):
    """
    Export inaFaceAnalyzer results to a video with incrusted faces bounding
    boxes and other analysis information

    Parameters
    ----------
    vid_src : str
        path to the input video.
    result_df : str or pandas.DataFrame
        path to the csv or pandas DataFrame corresponding to the analysis of the input video.
    vid_dst : str
        output path which will store the corresponding video.
        Must have MP4 extension
    analysis_fps: numeric or None
        The amount of frames per second which were analyzed
        if set to None, then consider that all video frames were processed
    """
    assert vid_dst[-4:].lower() == '.mp4', vid_dst


    video_props = get_video_properties(vid_src)
    fps, width, height = [video_props[e] for e in ['fps', 'width', 'height']]

    if analysis_fps is None:
        subsamp_coeff = 1
    else:
        subsamp_coeff = analysisFPS2subsamp_coeff(vid_src, analysis_fps)

    displaydf = _analysis2displaydf(result_df, fps, subsamp_coeff)

    font = cv2.FONT_HERSHEY_SIMPLEX

    with tempfile.TemporaryDirectory() as p:
        tmpout = '%s/%s.mp4' % (p, os.path.splitext(os.path.basename(vid_src))[0])

        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(tmpout, fourcc, fps / subsamp_coeff, (width, height))
        #print('out', out)

        for iframe, frame in video_iterator(vid_src, subsamp_coeff=subsamp_coeff):

            sdf = displaydf[displaydf.frame == iframe]

            for e in sdf.itertuples():

                x1, y1, x2, y2 = e.bbox
                color = _hex2rgb(e.rgb_color)
                cv2.putText(frame,e.text,(x1 - 100, y1 - 10 ), font, 0.7, color,2,cv2.LINE_AA)
                cv2.rectangle(frame, (x1, y1), (x2, y2), color, 8)

            ret = out.write(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))

        out.release()
        os.system('ffmpeg -y -i %s -i %s -map 0:v:0 -map 1:a? -vcodec libx264 -acodec copy %s' % (tmpout, vid_src, vid_dst))

try:
    # To be used only in ipython/collab environments
    from IPython.core.display import display
    from IPython.display import HTML
    from base64 import b64encode

    def notebook_display_remote_vid(fname, width=600):
        data = '<div align="middle"> <video width=%d controls> <source src=%s> </video></div>' % (width, fname)
        display(HTML(data))

    def notebook_display_local_vid(video_path, width = 600):
      # thanks https://androidkt.com/how-to-capture-and-play-video-in-google-colab/
      video_file = open(video_path, "r+b").read()
      video_url = f"data:video/mp4;base64,{b64encode(video_file).decode()}"
      return HTML(f"""<video width={width} controls><source src="{video_url}"></video>""")

except:
    pass


#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2021 Ina (David Doukhan & Zohra Rezgui - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pandas as pd
from abc import ABC, abstractmethod
from .opencv_utils import video_iterator, image_iterator, analysisFPS2subsamp_coeff
from .pyav_utils import video_keyframes_iterator
from .face_tracking import TrackerDetector
from .face_detector import LibFaceDetection, PrecomputedDetector
from .face_classifier import Resnet50FairFaceGRA
from .face_alignment import Dlib68FaceAlignment
from .face_preprocessing import preprocess_face


class FaceAnalyzer(ABC):
    """
    This is an abstract class containg the pipeline used to process
    images, videos, with/without tracking
    * image/video decoding
    * face detection
    * face tracking (optional)
    * eye detection
    * face preprocessing
    * face classification
    """

    # len of batches to be sent to face classifiers
    #batch_len = 32

    def __init__(self, face_detector = None, face_classifier = None, batch_len=32, verbose = False):
        """
        Constructor
        Parameters
        ----------
        face_detector : instance of face_detector.FaceDetector or None
            if None, LibFaceDetection is used by default
        face_classifier: instance of face_classifier.FaceClassifier or None
            if None, Resnet50FairFaceGRA is used by default (gender & age)
        verbose : boolean
            If True, will display several usefull intermediate images and results
        """

        # face detection system
        if face_detector is None:
            face_detector = LibFaceDetection()
        self.face_detector = face_detector

        # Face feature extractor from aligned and detected faces
        if face_classifier is None:
            face_classifier = Resnet50FairFaceGRA()
        self.classifier = face_classifier

        # if set to True, then the bounding box (manual or automatic) is set
        # to the smallest square containing the bounding box
        self.bbox2square = face_classifier.bbox2square

        # scaling factor to be applied to the face bounding box after detection
        # larger bounding box may help for sex classification from face
        self.bbox_scale = face_classifier.bbox_scale

        # face alignment module
        self.face_alignment = Dlib68FaceAlignment()

        # True if some verbose is required
        assert isinstance(verbose, bool)
        self.verbose = verbose

        # set to large values with large memory GPU for faster processing times !
        assert isinstance(batch_len, int) and batch_len > 0
        self.batch_len = batch_len


    @abstractmethod
    def __call__(self, src) :
        """
        Method to be implemented by each analyzer

        Parameters
        ----------
        src : str or list
            path to the video/image to be analyzed
            May also be a list of images
        Returns
        -------
        Results stored in a pandas DataFrame
        """
        pass

    def _process_stream(self, stream_iterator, detector):
        """
        Generic pipeline allowing to process image or video streams
        Faces are first detected, preprocessed and sent in batches in
        face classifiers

        Parameters
        ----------
        stream_iterator : iterator
            iterator returning decoded RBG images at each call together with
            an image identifier
            see: opencv_utils.video_iterator, opencv_utils.image_iterator,
            pyav_utils.video_keyframes_iterator
        detector : instance of face_tracking, tracker_detector or face_detector.FaceDetector

        Returns
        -------
        TYPE
            pandas Dataframe with analysis results

        """
        oshape = self.classifier.input_shape[:-1]

        lbatch_img = []
        linfo = []
        ldf = []

        # iterate on image list or video stream
        for iframe, frame in stream_iterator:

            # iterate on detected faces
            for detection in detector(frame, self.verbose):

                # preprocess detected faces: bbox normalization, eye detection, rotation, ...
                face_img, bbox = preprocess_face(frame, detection, self.bbox2square, self.bbox_scale, self.face_alignment, oshape, False)

                linfo.append([iframe, detection._replace(bbox=tuple(bbox))])
                # store preprocessed faces in a list, for further batch processing
                lbatch_img.append(face_img)

            # if enough faces were found, process a batch of faces
            while len(lbatch_img) > self.batch_len:
                df = self.classifier(lbatch_img[:self.batch_len], False, verbose=self.verbose)
                ldf.append(df)
                lbatch_img = lbatch_img[self.batch_len:]


        if len(lbatch_img) > 0:
            df = self.classifier(lbatch_img, False, verbose=self.verbose)
            ldf.append(df)

        if len(ldf) == 0:
            return pd.DataFrame(None, columns=(['frame'] + list(detector.output_type._fields) + self.classifier.output_cols))

        # return results as a pandas Dataframe
        df1 = pd.DataFrame({'frame' : [e[0] for e in linfo]})
        df2 = pd.DataFrame.from_records([e[1] for e in linfo], columns=detector.output_type._fields)
        if 'eyes' in df2.columns:
            df2 = df2.drop('eyes', axis=1)
        df3 = pd.concat(ldf).reset_index(drop=True)
        return pd.concat([df1, df2, df3], axis = 1)


class ImageAnalyzer(FaceAnalyzer):
    """
    ImageAnalyzer instances allow to detect and classify faces from images
    """
    def __call__(self, img_paths):
        """
        Parameters
        ----------
        img_paths : str or list
            path or list of paths to image file(s) to analyze
        Returns
        -------
        pandas Dataframe with column 'frame' containing the path to the source
        image. Remaining columns depend on processing options selected and
        contain bounding box, and face classification information

        """
        if isinstance(img_paths, str):
            stream = image_iterator([img_paths], verbose = self.verbose)
        else:
            stream = image_iterator(img_paths, verbose = self.verbose)
        return self._process_stream(stream, self.face_detector)


class VideoAnalyzer(FaceAnalyzer):
    """
    Video Analyzer allows to detect and classify faces in video streams
    """
    def __call__(self, video_path, fps = None,  offset = 0):
        """
        Pipeline function for face classification from videos (without tracking)

        Parameters:
            video_path: str
                Path to input video.
            fps: float or None (default)
                amount of video frames to process per seconds
                if set to None, all frames are processed (costly)
            offset: float (default: 0)
                Time in milliseconds to skip at the beginning of the video.

        Returns:
            Dataframe with frame and face information: frame position,
            coordinates, predictions, decision function,labels...
        """
        subsamp_coeff = 1 if fps is None else analysisFPS2subsamp_coeff(video_path, fps)
        stream = video_iterator(video_path, subsamp_coeff=subsamp_coeff, time_unit='ms', start=max(offset, 0), verbose=self.verbose)
        return self._process_stream(stream, self.face_detector)


class VideoKeyframes(FaceAnalyzer):
    """
    Face detection and analysis from video limited to video key frames
    https://en.wikipedia.org/wiki/Key_frame
    It allows to provide a video analysis summary in fast processing time, but
    with non uniform frame sampling rate
    """
    def __call__(self, video_path):

        """
        Pipeline function for face classification from videos, limited to key frames

        Parameters:
            video_path (string): Path for input video.
        Returns:
            Dataframe with frame and face information: frame position,
            coordinates, predictions, decision function,labels...
        """
        stream = video_keyframes_iterator(video_path, verbose=self.verbose)
        return self._process_stream(stream, self.face_detector)


class VideoTracking(FaceAnalyzer):
    """
    Video processing pipeline including face detection, tracking and classification
    Tracking is usually less costly than face detection (computation bottleneck)
    and allows to save computation time
    Classification decision functions and predictions are averaged for each
    tracked faces, allowing to obtain more robust analysis estimates
    """
    def __init__(self, detection_period, face_detector = None, face_classifier = None, batch_len=32, verbose = False):
        """
        Constructor

        Parameters
        ----------
        detection_period : int
            the face detection algorithm (costly) will be used once every
            'detection_period' analyzed frames.
            Ie: if set to 5, face detection will occur for 1/5 frames and the
            remaining 4/5 faces will be detected through a tracking procedure
                if set to 1: face detection will occur for each frame. Face
                tracking will also be used for each frames, since it will allow
                to group same faces under a person identifier
        face_detector : instance of face_detector.FaceDetector or None, optional
            if None, LibFaceDetection is used. The default is None.
        face_classifier : instance of face_classifier.FaceClassifier or None, optional
            if None, Resnet50FairFaceGRA is used (gender & age). The default is None.
        verbose : boolean, optional
            If True, will display several usefull intermediate images and results.
            The default is False.
        """
        super().__init__(face_detector, face_classifier, batch_len=batch_len, verbose=verbose)
        self.detection_period = detection_period

    def __call__(self, video_path, fps = None,  offset = 0):
        """
        Pipeline function for face classification from videos with tracking

        Parameters:
            video_path: str
                Path to input video.
            fps: float or None (default)
                amount of video frames to process per seconds
                if set to None, all frames are processed (costly)
            offset: float (default: 0)
                Time in milliseconds to skip at the beginning of the video.

        Returns:
            Dataframe with frame and face information: frame position,
            coordinates, predictions, decision function,labels...
            faceid column allow to keep track of each unique face found
            predictions and decision functions with '_avg' suffix are obtained
            through a smoothing procedure of decision functions for all faces
            with same faceid. Smoothed estimates are usually more robust than
            instantaneous ones
        """
        detector = TrackerDetector(self.face_detector, self.detection_period)

        subsamp_coeff = 1 if fps is None else analysisFPS2subsamp_coeff(video_path, fps)
        stream = video_iterator(video_path, subsamp_coeff=subsamp_coeff, time_unit='ms', start=max(offset, 0), verbose=self.verbose)

        df = self._process_stream(stream, detector)

        return self.classifier.average_results(df)



class VideoPrecomputedDetection(FaceAnalyzer):
    """
    Video analysis class to be used combined with pre-detected face bounding
    boxes (uncommon use-case)
    CANNOT BE USED WITH RESCALED OR SQUARIFIED BOUNDING BOXES!!!
    FACE ALIGNMENT REQUIRE TO HAVE ORIGINAL BOUNDING BOX TO PERFORM WELL!!
    """
    def __init__(self, face_classifier = None, verbose = False, bbox_scale=None, bbox2square=None):
        """
        Constructor

        Parameters
        ----------
        face_classifier: instance of face_classifier.FaceClassifier or None
            if None, Resnet50FairFaceGRA is used by default (gender & age)
        verbose : boolean
            If True, will display several usefull intermediate images and results
        bbox_scale: float or None
            scaling factor to be applied to the face bounding box after detection
            if not None, will overrides classifier's bbox scaling instructions
            usefull if provided bounding boxes are already transformed - default None
        bbox2square: boolean or None
            set bounding box to the smallest square containing the bounding box
            if not None, will ovverides classifier's bbox2square instructions
            usefull if provided bounding boxes are already transformed - default None
        """

        super().__init__(PrecomputedDetector(), face_classifier, verbose=verbose)
        if bbox_scale is not None:
            self.bbox_scale = bbox_scale
        if bbox2square is not None:
            self.bbox2square = bbox2square

    def __call__(self, video_path, lbbox, fps=None, start_frame = 0):
        """
        Pipeline function for face classification from videos using pre-detected faces

        Parameters:
            video_path: str
                Path to input video.
            lbbox: list of bounding boxes
                Each list element i contain either a tuple or a list of tuples
                (x1,y1,x2,y2) corresponding to the face found in the ith frame
            fps: float or None (default)
                amount of video frames to process per seconds
                if set to None, all frames are processed (costly)
            offset: float (default: 0)
                Time in milliseconds to skip at the beginning of the video.

        Returns:
            Dataframe with frame and face information: frame position,
            coordinates, predictions, decision function,labels...
        """
        detector = PrecomputedDetector(lbbox)

        subsamp_coeff = 1 if fps is None else analysisFPS2subsamp_coeff(video_path, fps)
        stream = video_iterator(video_path, subsamp_coeff=subsamp_coeff, time_unit='frame', start=start_frame, verbose=self.verbose)
        df = self._process_stream(stream, detector)

        assert len(detector.lbbox) == 0, 'the detection list is longer than the number of processed frames'
        return df



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import h5py
import numpy as np
from sklearn.svm import LinearSVC

# Load linear SVM serialized in a custom HDF5 format
# more robust than pickle
def svm_load(hdf5_fname):
    f = h5py.File(hdf5_fname, 'r')
    svm = LinearSVC()
    svm.classes_ = np.array(f['linearsvc/classes'][:]).astype('<U1')
    svm.intercept_ = f['linearsvc/intercept'][:]
    svm.coef_ = f['linearsvc/coef'][:]
    return svm

# TODO: test the load & dump pipeline
def svm_dump(hdf5_fname, model):
    with h5py.File(hdf5_fname, 'w') as fid:
        fid['linearsvc/intercept'] = model.intercept_
        fid['linearsvc/coef'] = model.coef_
        fid.create_dataset('linearsvc/classes', data = model.classes_.astype(h5py.special_dtype(vlen=str) ))



#!/usr/bin/env python
# encoding: utf-8

# This code is adapted from project libface detection
# https://github.com/ShiqiYu/libfacedetection
# libfacedetection is released under a 3 clause BSD license
# https://github.com/ShiqiYu/libfacedetection/blob/master/LICENSE

# To cite this work please refer to the 2 following papers :

# @article{facedetect-yu,
#  author={Yuantao Feng and Shiqi Yu and Hanyang Peng and Yan-ran Li and Jianguo Zhang}
#  title={Detect Faces Efficiently: A Survey and Evaluations},
#  journal={IEEE Transactions on Biometrics, Behavior, and Identity Science},
#  year={to appear}
#  }

# @article{eiou,
#  author={Peng, Hanyang and Yu, Shiqi},
#  journal={IEEE Transactions on Image Processing},
#  title={A Systematic IoU-Related Method: Beyond Simplified Regression for Better Localization},
#  year={2021},
#  volume={30},
#  pages={5032-5044},
#  doi={10.1109/TIP.2021.3077144}
#  }
#  https://ieeexplore.ieee.org/document/9429909

import sys
import numpy as np
from itertools import product
from typing import List, Tuple

class PriorBox(object):
    def __init__(self,
                 input_shape: Tuple[int] = (320, 240), # [in_w, in_h]
                 output_shape: Tuple[int] = (320, 240), # [img_w, img_h]
                 variance: List[float] = [0.1, 0.2]) -> None:
        super(PriorBox, self).__init__()

        self.min_sizes = [[10, 16, 24], [32, 48], [64, 96], [128, 192, 256]]
        self.steps = [8, 16, 32, 64]

        self.in_w, self.in_h = input_shape
        self.out_w, self.out_h = output_shape
        self.variance = variance

        for ii in range(4):
            if(self.steps[ii] != pow(2,(ii+3))):
                print("steps must be [8,16,32,64]")
                sys.exit()

        self.feature_map_2th = [int(int((self.in_h + 1) / 2) / 2),
                                int(int((self.in_w + 1) / 2) / 2)]
        self.feature_map_3th = [int(self.feature_map_2th[0] / 2),
                                int(self.feature_map_2th[1] / 2)]
        self.feature_map_4th = [int(self.feature_map_3th[0] / 2),
                                int(self.feature_map_3th[1] / 2)]
        self.feature_map_5th = [int(self.feature_map_4th[0] / 2),
                                int(self.feature_map_4th[1] / 2)]
        self.feature_map_6th = [int(self.feature_map_5th[0] / 2),
                                int(self.feature_map_5th[1] / 2)]

        self.feature_maps = [self.feature_map_3th, self.feature_map_4th,
                             self.feature_map_5th, self.feature_map_6th]

        self.priors = self.generate_priors()

    def generate_priors(self) -> np.ndarray:
        anchors = np.empty(shape=[0, 4])
        for k, f in enumerate(self.feature_maps):
            min_sizes = self.min_sizes[k]
            for i, j in product(range(f[0]), range(f[1])): # i->in_h, j->in_w
                for min_size in min_sizes:
                    s_kx = min_size / self.in_w
                    s_ky = min_size / self.in_h

                    cx = (j + 0.5) * self.steps[k] / self.in_w
                    cy = (i + 0.5) * self.steps[k] / self.in_h

                    anchors = np.vstack(
                        (anchors, np.array([cx, cy, s_kx, s_ky]))
                    )
        return anchors

    def decode(self, loc: np.ndarray, conf: np.ndarray, iou: np.ndarray, ignore_score: float = 0.6) -> np.ndarray:
        '''Decodes the locations (x1, y1, x2, y2) and scores (c) from the priors, and the given loc and conf.
        Ignore low scores based on ignore_score.
        Args:
            loc (np.ndarray): loc produced from loc layers of shape [num_priors, 4]. '4' for [x_c, y_c, w, h].
            conf (np.ndarray): conf produced from conf layers of shape [num_priors, 2]. '2' for [p_non_face, p_face].
            iou (np.ndarray): iou produced from iou layers of shape [num_priors, 1]. '1' for [iou].
            ignore_score (float): used to filter out low score instances.
        Return:
            dets (np.ndarray): dets is concatenated by bboxes, landmarks and scores.
                bboxes consists of num_priors * (x1, y1, x2, y2) of shape [num_priors, 4].
                landmarks consists of num_priors * (x_le, y_le, x_re, y_r2, x_n, y_n, x_ml, y_ml, x_mr, y_mr) of shape [num_priors, 5*2].
        '''
        # get score
        cls_scores = conf[:, 1]
        iou_scores = iou[:, 0]
        # clamp
        _idx = np.where(iou_scores < 0.)
        iou_scores[_idx] = 0.
        _idx = np.where(iou_scores > 1.)
        iou_scores[_idx] = 1.
        scores = np.sqrt(cls_scores * iou_scores)
        scores = scores[:, np.newaxis]

        # ignore low scores
        idx = np.where(scores[:, 0] > ignore_score)[0]
        scores = scores[idx]
        loc = loc[idx]
        priors = self.priors[idx]

        # get bboxes
        bboxes = np.hstack((
            priors[:, 0:2]+loc[:, 0:2]*self.variance[0]*priors[:, 2:4],
            priors[:, 2:4]*np.exp(loc[:, 2:4]*self.variance)
        ))
        # (x_c, y_c, w, h) -> (x1, y1, w, h)
        bboxes[:, 0:2] -= bboxes[:, 2:4] / 2
        # scale recover
        bbox_scale = np.array([self.out_w, self.out_h]*2)
        bboxes = bboxes * bbox_scale

        # get landmarks
        landmarks = np.hstack((
            priors[:, 0:2]+loc[:,  4: 6]*self.variance[0]*priors[:, 2:4],
            priors[:, 0:2]+loc[:,  6: 8]*self.variance[0]*priors[:, 2:4],
            priors[:, 0:2]+loc[:,  8:10]*self.variance[0]*priors[:, 2:4],
            priors[:, 0:2]+loc[:, 10:12]*self.variance[0]*priors[:, 2:4],
            priors[:, 0:2]+loc[:, 12:14]*self.variance[0]*priors[:, 2:4]
        ))
        # scale recover
        landmark_scale = np.array([self.out_w, self.out_h]*5)
        landmarks = landmarks * landmark_scale

        dets = np.hstack((bboxes, landmarks, scores))
        return dets



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan & Thomas Petit - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This code is a modification of nlhkh's project face-alignment-dlib
# See github.com/nlhkh/face-alignment-dlib/blob/master/utils.py
# It has been adapted by Thomas PETIT (github.com/w2ex) & David Doukhan

import dlib
from .rect import Rect
from .remote_utils import get_remote

def _extract_eye(shape, eye_indices):
    points = map(lambda i: shape.part(i), eye_indices)
    return list(points)


def _extract_eye_center(shape, eye_indices):
    points = _extract_eye(shape, eye_indices)
    xs = map(lambda p: p.x, points)
    ys = map(lambda p: p.y, points)
    return sum(xs) // 6, sum(ys) // 6


def extract_left_eye_center(shape):
    LEFT_EYE_INDICES = [36, 37, 38, 39, 40, 41]
    return _extract_eye_center(shape, LEFT_EYE_INDICES)


def extract_right_eye_center(shape):
    RIGHT_EYE_INDICES = [42, 43, 44, 45, 46, 47]
    return _extract_eye_center(shape, RIGHT_EYE_INDICES)

class Dlib68FaceAlignment:
    """
    Methods for detecting eye centers based on Dlib's set of 68 facial landmarks
    this information may be used to rotate the input image such as the eyes lie
    on a horizontal line
    Aligned faces allow to obtain better results on face classication tasks
    """

    def __init__(self, verbose=False):
        """
        Load dlib's 68 facial landmark detection model

        Parameters
        ----------
        verbose : boolean, optional. The default is False.
            If set to True, resulting rotated image will be displayed.
        """
        self.model = dlib.shape_predictor(get_remote('shape_predictor_68_face_landmarks.dat'))
        self.verbose = verbose

    def __call__(self, frame, bb):
        """
        Detects left and right eye centers

        Parameters
        ----------
        frame : numpy.ndarray (height,with, 3)
            RGB image data
        bb : (x1, y1, x2, y2) or None
            Location of the face in the frame
            If set to None, the whole image is considered as a face

        Returns
        -------
        left_eye : (x,y)
            Center of the left eye in the input image
        right_eye : (x,y)
            Center of the right eye in the input image

        """
        if bb is None:
            bb = Rect(0, 0, frame.shape[1], frame.shape[0])
        shape = self.model(frame, bb.to_dlibInt())
        left_eye = extract_left_eye_center(shape)
        right_eye = extract_right_eye_center(shape)
        return left_eye, right_eye



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019-2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.# -*- coding: utf-8 -*-




import Pyro4
import sys
import os
import socket
from collections import namedtuple
import pandas as pd
import inaFaceAnalyzer.commandline_utils as ifacu
from inaFaceAnalyzer.display_utils import ass_subtitle_export, video_export


description = '''Worker in charge of analyzing documents and receive orders
from a central job server. Workers need to have access to input files (url or path)
and must be allowed to write results in a centralized storage directory
To be used jointly with ina_face_analyzer_distributed_server.py .
The server must be run before runing the worker.
'''

hserver_uri = '''
The URI displayed by the server to be copy/pasted after runing program
ina_face_analyzer_distributed_server.py  .
Sample URI: PYRO:obj_25e3ee9d312848fcaf0784c2b80933c4@blahtop:44175
'''


if __name__ == '__main__':

    # parse command line arguments
    parser = ifacu.new_parser(description)
    parser.add_argument(dest='server_uri', help=hserver_uri)
    ifacu.add_batchsize(parser)
    args = parser.parse_args()

    # setup network configuration
    hostname = socket.gethostname()
    jobserver = Pyro4.Proxy(args.server_uri)

    # get processing settings from server, and update with command line arguments
    server_args = jobserver.get_analysis_args('initializing ' + hostname)
    print('recieved args', server_args)
    server_args['batch_size'] = args.batch_size
    nt = namedtuple('inaFaceAnalyzerArgs', server_args)
    args = nt(**server_args)

    # to be used at engine inference
    dargs = {}
    if args.fps:
        dargs = {'fps': args.fps}

    ## perform engine instantiation
    engine = ifacu.engine_factory(args)


    ret = 'first call'
    stopit = False
    while not stopit:
        try:
            src, dst, dst_ass, dst_mp4 = jobserver.get_job('%s %s' % (hostname, ret))
        except StopIteration:
            print('all jobs are done')
            stopit = True
            sys.exit(0)

        print('received job', src, dst, dst_ass, dst_mp4)
        df = None

        ret = ''
        if not os.path.exists(dst):
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            df = engine(src, **dargs)
            df.to_csv(dst, index=False)
            ret += '%s done ' % dst

        if dst_ass == dst_ass and not os.path.exists(dst_ass):
            os.makedirs(os.path.dirname(dst_ass), exist_ok=True)
            if df is None:
                df = pd.read_csv(dst)
            ass_subtitle_export(src, df, dst_ass, analysis_fps=args.fps)
            ret += '%s done ' % dst_ass

        if dst_mp4 == dst_mp4 and not os.path.exists(dst_mp4):
            os.makedirs(os.path.dirname(dst_mp4), exist_ok=True)
            if df is None:
                df = pd.read_csv(dst)
            video_export(src, df, dst_mp4, analysis_fps=args.fps)
            ret += '%s done ' % dst_mp4


#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import av
import av.datasets
from .opencv_utils import disp_frame

def video_keyframes_iterator(video_path, verbose=False):

    content = av.datasets.curated(os.path.abspath(video_path))
    with av.open(content) as container:
        # Signal that we only want to look at keyframes.
        stream = container.streams.video[0]
        stream.codec_context.skip_frame = 'NONKEY'
    
        for frame in container.decode(stream):
            ndframe = frame.to_rgb().to_ndarray()
            iframe = int(frame.time/frame.time_base)
            if verbose:
                print('frame', iframe)
                disp_frame(ndframe)
            yield iframe, ndframe



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import cv2
from inaFaceAnalyzer import ImageAnalyzer
from inaFaceAnalyzer.face_classifier import Resnet50FairFaceGRA
import inaFaceAnalyzer.commandline_utils as ifacu
from inaFaceAnalyzer.face_detector import facedetection_cmdlineparser, facedetection_factory


### BUILD PARSER
description = 'Webcam demo. Funny and Real-Time'
parser = ifacu.new_parser(description)
# update parser with face detection arguments
parser.add_argument("-h", "--help", action="help", help="show this help message and exit")
facedetection_cmdlineparser(parser)

# PARSE ARGUMENTS
args = parser.parse_args()
detector = facedetection_factory(args)


gi = ImageAnalyzer(face_classifier=Resnet50FairFaceGRA(), face_detector=detector)

font = cv2.FONT_HERSHEY_SIMPLEX


video_capture = cv2.VideoCapture(0)


iframe = 0

while True:
    # Capture frame-by-frame
    ret, frame = video_capture.read()

    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

    faces = gi._process_stream([(iframe, frame)], detector)

    # Draw a rectangle around the faces
    for e in faces.itertuples():
        x1, y1, x2, y2 = e.bbox

        text = 'sex: %s - %.1f; age: %.1f' % (e.sex_label, e.sex_decfunc, e.age_label)
        if e.sex_label == 'm': # blue
            color = (0,0,255)
        else: # red
            color = (255,0,0)
        cv2.putText(frame,text,(x1 - 100, y1 - 10 ), font, 0.7, color,2,cv2.LINE_AA)

        cv2.rectangle(frame, (x1, y1), (x2, y2), color, 2)

    # Display the resulting frame
    cv2.imshow('Video', frame)

    if cv2.waitKey(1) & 0xFF == ord('q'):
        break

    iframe += 1

# When everything is done, release the capture
video_capture.release()
cv2.destroyAllWindows()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan & Zohra Rezgui- http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from abc import ABC, abstractmethod
from typing import NamedTuple
import cv2
import numpy as np
import onnxruntime

from .rect import Rect
from .remote_utils import get_remote
from .opencv_utils import disp_frame_shapes, disp_frame
from .libfacedetection_priorbox import PriorBox


class Detection(NamedTuple):
    bbox : Rect
    detect_conf : float

class DetectionEyes(NamedTuple):
    """
    Contains a detection (Rect bounding box & detection confidence)
    + eyes coordinates (x1, y1, x2, y2) for left eye and right eye
    """
    bbox : Rect
    detect_conf : float
    eyes : Rect


class FaceDetector(ABC):
    def __init__(self, minconf, min_size_px, min_size_prct, padd_prct):
        self.minconf = minconf
        self.min_size_px = min_size_px
        self.min_size_prct = min_size_prct
        self.padd_prct = padd_prct

    def __call__(self, frame, verbose=False):

        tmpframe = frame

        if self.padd_prct:
            tmpframe, yoffset, xoffset = _blackpadd(frame, self.padd_prct)

        lret = self._call_imp(tmpframe)


        # filter detected faces to return only faces with a dimension length
        # (absolute or relative)
        # face classification algorithms may be affected by small face sizes
        min_frame_dim = min(frame.shape[:2])
        min_face_size = max(self.min_size_px, self.min_size_prct * min_frame_dim)
        if min_face_size > 0:
            lret = [e for e in lret if e.bbox.max_dim_len >= min_face_size]

        if self.padd_prct:
            lret = [e._replace(bbox=e.bbox.transpose(-xoffset, -yoffset)) for e in lret]

        if verbose:
            print('%d DETECTED FACES' % len(lret))
            disp_frame_shapes(frame, [e.bbox for e in lret], [])
            for detection in lret:
                x1, y1, x2, y2 = [e for e in detection.bbox.to_int()]
                print(detection)
                x1 = max(x1, 0)
                y1 = max(y1, 0)
                x2 = min(x2, frame.shape[1])
                y2 = min(y2, frame.shape[0])
                disp_frame(frame[y1:y2, x1:x2, :])


        return lret

    @classmethod
    @abstractmethod
    def output_type() : pass

    @abstractmethod
    def _call_imp(self, frame): pass

    def most_central_face(self, frame, contain_center=True, verbose=False):
        """
        This method returns the detected face which is closest from the center
        if contain_center == True, the returned face MUST include image center
        Usefull for preprocessign face datasets containing several faces per image
        Parameters
        ----------
        frame : nd.array (height, width, 3)
            RGB image data.
        verbose : boolean, optional
            Display detected faces. The default is False.
        Returns
        -------
        A Detection named tuple if a face maching the conditions has been detected else None
        """

        frame_center = (frame.shape[1] / 2, frame.shape[0] / 2)

        # keep faces containing image center
        if contain_center:
            faces = [f for f in self(frame, verbose) if frame_center in f.bbox]
        else:
            faces = [f for f in self(frame, verbose)]

        if len(faces) == 0:
            return None

        ldists = [_sqdist(f.bbox.center, frame_center) for f in faces]
        am = np.argmin(ldists)

        return faces[am]


    def get_closest_face(self, frame, ref_bbox, min_iou=.7, squarify=True, verbose=False):
        """
        Some face corpora may contain pictures with several faces
        together with the reference bounding box of annotated faces
        This function is aimed at preprocessing such face corpora
        Automatic face detection is used, and the detected face with the largest
        iou with the reference bounding box is returned
        if no detected face corresponds to this IOU criteria, returns None
        """
        if not isinstance(ref_bbox, Rect):
            ref_bbox = Rect(*ref_bbox)

        # get closest detected faces from ref_bbox
        if squarify:
            f = lambda x: x.square
        else:
            f = lambda x: x

        ref_bbox = f(ref_bbox)

        lfaces = self(frame, verbose)
        if len(lfaces) == 0:
            return None

        liou = [f(ref_bbox).iou(f(detection.bbox)) for detection in lfaces]


        if verbose:
            print([f(detection.bbox) for detection in lfaces])
            print('liou', liou)
        am = np.argmax(liou)
        if liou[am] < min_iou:
            return None
        return lfaces[am]



def _sqdist(p1, p2):
    '''
    return squared distance between points p1(x,y) and p2(x,y)
    '''
    return (p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2

def _blackpadd(frame, paddpercent):
    # add black around image
    y, x, z = frame.shape
    #offset = int(max(x, y) * paddpercent)
    xoffset = int(x * paddpercent)
    yoffset = int(y * paddpercent)
    ret = np.zeros((y + 2 * yoffset, x + 2 * xoffset, z), dtype=frame.dtype)
    ret[yoffset:(y + yoffset), xoffset:(x + xoffset), :] = frame
    return ret, yoffset, xoffset



class OcvCnnFacedetector(FaceDetector):
    """
    opencv default CNN face detector
    """
    output_type = Detection

    def __init__(self, minconf=0.65, min_size_px=30, min_size_prct=0, padd_prct=0.15):
        """
        Parameters
        ----------
        minconf : float, optional
           minimal face detection confidence. The default is 0.65.
        paddpercent : float, optional
            input frame is copy passted within a black image with black pixel
            padding. the resulting dimensions is width * (1+2*paddpercent)

        """
        super().__init__(minconf, min_size_px, min_size_prct, padd_prct)

        fpb = get_remote('opencv_face_detector_uint8.pb')
        fpbtxt = get_remote('opencv_face_detector.pbtxt')
        self.model = cv2.dnn.readNetFromTensorflow(fpb, fpbtxt)


    def _call_imp(self, frame):
        """
        Detect faces from an image

        Parameters:
            frame (array): Image to detect faces from.

        Returns:
            faces_data (list) : List containing :
                                - the bounding box
                                - face detection confidence score
        """

        faces_data = []
        h, w, z = frame.shape

        # The CNN is intended to work images resized to 300*300
        # tests were carried on using different input size and were associated
        # to usatisfactory results
        blob = cv2.dnn.blobFromImage(frame, 1.0, (300, 300), [104, 117, 123], True, False)
        self.model.setInput(blob)
        detections = self.model.forward()

        assert(np.all(-np.sort(-detections[:,:,:,2]) == detections[:,:,:,2]))

        for i in range(detections.shape[2]):
            confidence = detections[0, 0, i, 2]

            if confidence < self.minconf:
                break

            bbox = Rect(*detections[0, 0, i, 3:7])
            # remove noisy detections coordinates
            if bbox.x1 >= 1 or bbox.y1 >= 1 or bbox.x2 <= 0 or bbox.y2 <= 0:
                continue
            if bbox.x1 >= bbox.x2 or bbox.y1 >= bbox.y2:
                continue

            # Map relative coordinates 0...1 to absolute  frame width and height
            bbox = bbox.mult(w, h)
            faces_data.append(Detection(bbox, confidence))

        return faces_data


class IdentityFaceDetector(FaceDetector):
    output_type = Detection
    def __init__(self):
        super().__init__(0, 0, 0, 0)
    def _call_imp(self, frame):
        return [Detection(Rect(0, 0, frame.shape[1], frame.shape[0]), np.NAN)]

class LibFaceDetection(FaceDetector):
    """
    This class wraps the detection model provided in libfacedetection
    See: https://github.com/ShiqiYu/libfacedetection
    """

    output_type = DetectionEyes

    def __init__(self, minconf=.98, min_size_px=30, min_size_prct=0, padd_prct=0):
        super().__init__(minconf, min_size_px, min_size_prct, padd_prct)
        model_src = get_remote('libfacedetection-yunet.onnx')
        try:
            self.model = onnxruntime.InferenceSession(model_src, providers=['CUDAExecutionProvider', 'CPUExecutionProvider'])
        except:
            self.model = onnxruntime.InferenceSession(model_src, providers=['CPUExecutionProvider'])
        self.nms_thresh = 0.3 # Threshold for non-max suppression
        self.keep_top_k = 750 # Keep keep_top_k for results outputing
        self.dprior = {}

    def _call_imp(self, frame):
        bgr_frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        h, w, _ = frame.shape

        # convert to NN input
        blob = np.expand_dims(np.transpose(bgr_frame, (2, 0, 1)), axis = 0).astype(np.float32)
        # NN inference
        loc, conf, iou = self.model.run([], {'input': blob})

        # Decode bboxes and landmarks
        # TODO: set a limit of dict length ? There may be RAM issues when
        # considering a large image collection with heterogenous sizes
        if (w, h) not in self.dprior:
            self.dprior[(w, h)] = PriorBox(input_shape=(w, h), output_shape=(w, h))
        pb = self.dprior[(w, h)]

        dets = pb.decode(loc, conf, iou, self.minconf)

        # dirty hack used for google collab compatibility
        if len(dets.shape) == 3 and dets.shape[1] == 1:
            dets = dets.reshape((dets.shape[0], dets.shape[2]))
        assert len(dets.shape) == 2, dets.shape

        # NMS
        if dets.shape[0] > 0:
             # NMS from OpenCV
             keep_idx = cv2.dnn.NMSBoxes(
                  bboxes=dets[:, 0:4].tolist(),
                  scores=dets[:, -1].tolist(),
                  score_threshold=self.minconf,
                  nms_threshold=self.nms_thresh,
                  eta=1,
                  top_k=self.keep_top_k)
             dets = dets[keep_idx]
        else:
            return []

        # dirty hack used for google collab compatibility
        # it works - to be investiguated
        if len(dets.shape) == 3 and dets.shape[1] == 1:
            dets = dets.reshape((dets.shape[0], dets.shape[2]))
        assert len(dets.shape) == 2, dets.shape
        assert dets.shape[1] == 15, dets.shape

        lret = []
        for i in range(len(dets)):
            score = dets[i,-1]
            x1, y1, w, h = dets[i,:4]
            bbox = Rect(x1, y1, x1 + w, y1 + h)
            eyes = Rect(*dets[i, 4:8])
            lret.append(DetectionEyes(bbox, score, eyes))

        return lret

class PrecomputedDetector(FaceDetector):
    output_type = Detection
    def __init__(self, lbbox = []):
        super().__init__(0, 0, 0, 0)
        self.lbbox = lbbox.copy()
    def _call_imp(self, frame):
        if len(self.lbbox) == 0:
            return []
        ret = self.lbbox.pop(0)
        if isinstance(ret, tuple):
            ret = [ret]
        return [Detection(Rect(*e), None) for e in ret]


def facedetection_cmdline(parser):
    '''
    Update command line parser with face detection related arguments
    Parameters
    ----------
    parser : argparse.ArgumentParser
        command line parser to be updated

    '''
    da = parser.add_argument_group('optional arguments related to face detection')

    da.add_argument ('--face_detector', default='LibFaceDetection',
                     choices=['LibFaceDetection', 'OcvCnnFacedetector'],
                     help='''face detection module to be used:
                         LibFaceDetection can take advantage of GPU acceleration and has a higher recall.
                         OcvCnnFacedetector is embedded in OpenCV. It is faster for large resolutions since it first resize input frames to 300*300. It may miss small faces''')

    da.add_argument('--face_detection_confidence', type=float,
                    help='''minimal confidence threshold to be used for face detection.
                        Default values are 0.98 for LibFaceDetection and 0.65 for OcvCnnFacedetector''')


    da.add_argument('--min_face_size_px', default=30, type=int, dest='size_px',
                    help='''minimal absolute size in pixels of the faces to be considered for the analysis.
                    Optimal classification results are obtained for sizes above 75 pixels.''')

    da.add_argument('--min_face_size_percent', default=0, type=float, dest='size_prct',
                    help='''minimal relative size (percentage between 0 and 1) of the
                    faces to be considered for the analysis with respect to image frames
                    minimal dimension (generally height for videos)''')

    da.add_argument('--face_detection_padding', default=None, type=float, dest='face_detection_padding',
                    help='''Black padding percentage to be applied to image frames before face detection.
                    0.15 Padding may help detecting large faces occupying the whole image with OcvCnnFacedetector.
                    Default padding values are 0.15 for OcvCnnFacedetector and 0 for LibFaceDetection''')

def facedetection_factory(args):
    '''
    Instanciate a face detection object from parsed command line arguments

    Parameters
    ----------
    args : Namespace
        Namespace containing fields face_detector, face_detection_confidence,
        min_face_size_px, min_face_size_percent

    Returns
    -------
    instance of class FaceDetector

    '''
    dargs = {'min_size_px': args.size_px, 'min_size_prct': args.size_prct}
    if args.face_detection_padding is not None:
        dargs['padd_prct'] = args.face_detection_padding
    if args.face_detection_confidence:
        dargs['minconf'] = args.face_detection_confidence

    if args.face_detector == 'LibFaceDetection':
        detector = LibFaceDetection(**dargs)
    elif args.face_detector == 'OcvCnnFacedetector':
        detector = OcvCnnFacedetector(**dargs)
    else:
        raise NotImplementedError(args.face_dector)

    return detector



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2019 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import os
from setuptools import setup #, find_packages
import versioneer

KEYWORDS = '''
gender-equality
gender-classification'''.strip().split('\n')

CLASSIFIERS=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
#    'Programming Language :: Python :: 3.10', # dependency onnxruntime-gpu does not support Python 3.10 yet
    'Topic :: Multimedia :: Video',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Image Processing',
    'Topic :: Scientific/Engineering :: Image Recognition',
    'Topic :: Sociology',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Sociology',
]

DESCRIPTION='inaFaceAnalyzer is a Python toolbox for large-scale face-based \
analysis of image and video streams. It provides fast API and command line programs allowing to perform \
face detection, face tracking, gender and age prediction, and export to CSV or rich ASS subtitles'

# read the contents of your README file
with open('README.md', 'r') as fid:
    long_description = fid.read()


setup(
    name = "inaFaceAnalyzer",
    version = versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author = "David Doukhan, Zohra Rezgui, Thomas Petit",
    author_email = "david.doukhan@gmail.com, zohra.rzg@gmail.com, tpetit@ina.fr",
    test_suite="test_inaFaceAnalyzer.py",
    description = DESCRIPTION,
    license = "MIT",
    install_requires=['opencv-contrib-python', 'dlib', 'pandas', 'sklearn',
                      'h5py', 'matplotlib', 'onnxruntime-gpu', 'cheetah3',
                      'av', 'tensorflow>=2.6', 'pyro4'],
    url = "https://github.com/ina-foss/inaFaceAnalyzer",
    packages=['inaFaceAnalyzer'],
    keywords = KEYWORDS,
    #packages = find_packages(),
    include_package_data = True,
    data_files = ['LICENSE'], #'inaFaceAnalyzer/template.ass'],
    long_description = long_description,
    long_description_content_type='text/markdown',
    scripts=[os.path.join('scripts', script) for script in \
             ['ina_face_analyzer.py', 'ina_face_analyzer_webcam_demo.py',
              'ina_face_analyzer_distributed_server.py',
              'ina_face_analyzer_distributed_worker.py']],
    classifiers=CLASSIFIERS,
    python_requires='>=3.7',

)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License

# Copyright (c) 2021 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import cv2
import numpy as np
import pylab as plt

def video_iterator(src, time_unit='frame', start=None, stop=None, subsamp_coeff=1, verbose=False):

    # cv2.CAP_PROP_POS_MSEC property was not used because it is buggy

    cap = cv2.VideoCapture(src)

    if not cap.isOpened():
        raise Exception("Video file %s does not exist or is invalid" % src)

    unit = cv2.CAP_PROP_POS_FRAMES

    if time_unit == 'frame':
        pass
    elif time_unit == 'ms':
        fps = cap.get(cv2.CAP_PROP_FPS)
        if start is not None:
            start = int(start * fps // 1000)
        if stop is not None:
            stop = int(stop * fps // 1000)
    else:
        raise NotImplementedError

    if start is not None:
        cap.set(unit, start)

    while cap.isOpened():

        if stop is not None and cap.get(unit) > stop:
            break

        ret, frame = cap.read()
        if not ret:
            break

        # skip frames for subsampling reasons
        iframe = cap.get(cv2.CAP_PROP_POS_FRAMES) - 1
        if iframe % subsamp_coeff != 0:
            continue


        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

        if verbose:
            print('frame ', iframe)
            plt.imshow(frame)
            plt.show()

        yield int(iframe), frame

    cap.release()

def image_iterator(lfiles, verbose = False):
    for f in lfiles:
        yield f, imread_rgb(f, verbose)

def get_video_properties(src):
    cap = cv2.VideoCapture(src)

    if not cap.isOpened():
        raise Exception("Video file %s does not exist or is invalid" % src)

    dret = {}
    dret['fps'] = int(cap.get(cv2.CAP_PROP_FPS))
    dret['width'] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    dret ['height'] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    cap.release()
    return dret

def analysisFPS2subsamp_coeff(vid_src, analysisFPS):
    props = get_video_properties(vid_src)
    return int(np.round(props['fps'] / analysisFPS))


def disp_frame(frame):
    plt.imshow(frame)
    plt.show()
    return frame

def disp_frame_shapes(frame, lrect = [], lpoint = []):
    tmpframe = frame.copy()
    for bbox in lrect:
        print('bbox', bbox)
        bbox = [int(e) for e in bbox]
        cv2.rectangle(tmpframe, (bbox[0], bbox[1]), (bbox[2], bbox[3]), (0, 255, 0), 8)
    for point in lpoint:
        print('point', point)
        point = [int(e) for e in point]
        cv2.circle(tmpframe, point, radius=4, color=(0, 0, 255), thickness=-1)
    return disp_frame(tmpframe)

def imwrite_rgb(dst, img):
    img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    if not cv2.imwrite(dst, img):
        raise Exception('cannot write image %s' % dst)

def imread_rgb(img_path, verbose=False):
    img = cv2.imread(img_path)
    frame = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    if verbose:
        print('raw image ' + img_path)
        plt.imshow(frame)
        plt.show()
    return frame



#!/usr/bin/env python
# encoding: utf-8
#
# !!! WARNING
#
# This code contains patches for python module keras-vggface
# keras-vggface is a python module created by Refikcanmalli
# it is released under MIT LICENSE and available at
# https://github.com/rcmalli/keras-vggface
#
# The module is no more maintained and the code bellow is a minimal copy/pasted
# adaptation allowing to use the VGG16 model provided in this package
# 
# In order to use a fully patched version of keras-vggface, see this fork
# https://github.com/DavidDoukhan/keras-vggface

# The MIT License

# Copyright (c) 2019-2022 Ina (David Doukhan - http://www.ina.fr/)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from tensorflow.keras.layers import Input, GlobalAveragePooling2D, Conv2D, MaxPooling2D
from tensorflow.python.keras.utils.data_utils import get_file
from tensorflow.keras.models import Model

import numpy as np

def VGG16(input_shape):
    img_input = Input(shape=input_shape)

    # Block 1
    x = Conv2D(64, (3, 3), activation='relu', padding='same', name='conv1_1')(
        img_input)
    x = Conv2D(64, (3, 3), activation='relu', padding='same', name='conv1_2')(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='pool1')(x)

    # Block 2
    x = Conv2D(128, (3, 3), activation='relu', padding='same', name='conv2_1')(
        x)
    x = Conv2D(128, (3, 3), activation='relu', padding='same', name='conv2_2')(
        x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='pool2')(x)

    # Block 3
    x = Conv2D(256, (3, 3), activation='relu', padding='same', name='conv3_1')(
        x)
    x = Conv2D(256, (3, 3), activation='relu', padding='same', name='conv3_2')(
        x)
    x = Conv2D(256, (3, 3), activation='relu', padding='same', name='conv3_3')(
        x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='pool3')(x)

    # Block 4
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='conv4_1')(
        x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='conv4_2')(
        x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='conv4_3')(
        x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='pool4')(x)

    # Block 5
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='conv5_1')(
        x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='conv5_2')(
        x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='conv5_3')(
        x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='pool5')(x)

    x = GlobalAveragePooling2D()(x)


    inputs = img_input
    # Create model.
    model = Model(inputs, x, name='vggface_vgg16')  # load weights

    url = 'https://github.com/rcmalli/keras-vggface/releases/download/v2.0/rcmalli_vggface_tf_notop_vgg16.h5'
    weights_path = get_file('rcmalli_vggface_tf_notop_vgg16.h5',
                            url , cache_subdir='models/vggface')
    
    model.load_weights(weights_path, by_name=True)

    return model


def preprocess_input(x):
    x_temp = np.copy(x)

    x_temp = x_temp[..., ::-1]
    x_temp[..., 0] -= 93.5940
    x_temp[..., 1] -= 104.7624
    x_temp[..., 2] -= 129.1863

    return x_temp